import Graphics.Gloss
import Graphics.Gloss.Juicy ( loadJuicyPNG )
import Graphics.Gloss.Interface.Pure.Game ( Event )
import Graphics.Gloss.Data.Vector ()
import Graphics.Rendering.OpenGL (patchVertices)
import Control.Exception (try, SomeException)
import Data.Maybe (catMaybes)
import Data.Either (either)

backgroundPath :: FilePath
backgroundPath = "D:/onedriveschool/OneDrive - University of Edinburgh/Gradeone/Intro to computation/Compepition/Background2.png"

window :: Display
window = InWindow "Solar System" (1920, 1080) (10, 10)
globalScaleConstant :: Float
globalScaleConstant = 0.005   -- global constants
earthDiameter = 1

imagePaths :: [FilePath]
imagePaths = ["D:/onedriveschool/OneDrive - University of Edinburgh/Gradeone/Intro to computation/Compepition/" ++ p |
    p <- ["Sun.png", "Mercury.png", "Venus.png", "Earth.png", "Mars.png", "Jupiter.png", "Saturn.png", "Uranus.png", "Neptune.png"]]

-- Earth 's scale factor
earthOrbitRadiusPixels = 80.0
earthOrbitRadiusAU = 1.0  
scaleFactor = earthOrbitRadiusPixels / earthOrbitRadiusAU

originalSunDiameterPixels = 159.5

--Sun' scale daimeter
scaledSunDiameter = originalSunDiameterPixels * globalScaleConstant * scaleFactor
orbitOffset = scaledSunDiameter / 100

orbitRadii = [90, 120, 150, 200, 320, 470, 620, 770]    --The scaling of the orbits of the eight major planets is not proportional due to the limited screen space.
scaledOrbitRadii = map (\au -> (au * scaleFactor) + orbitOffset) orbitRadii

loadPlanetPicture :: FilePath -> IO (Maybe Picture)
loadPlanetPicture path = do
  result <- try (loadJuicyPNG path) :: IO (Either SomeException (Maybe Picture))
  return $ either (const Nothing) id result

initPlanetData :: IO [PlanetData]
initPlanetData = do
  pics <- mapM loadPlanetPicture imagePaths
  let planetTypes = [Sun, Mercury, Venus, Earth, Mars, Jupiter, Saturn, Uranus, Neptune]
  let angularVelocities = map (\period -> 2 * pi / period) [0.6025, 1.5375, 2.5, 4.7025, 29.65, 73.65, 60.025, 412]
  return $ zip3 pics planetTypes (zip scaledOrbitRadii angularVelocities)

main :: IO ()
main = do
  maybeBackground <- loadJuicyPNG backgroundPath
  planetData <- planetDataIO
  let initialState = (0, planetData)
  play window black 30 initialState (render maybeBackground) ignoreEvents update

planetTypes :: [PlanetType]
planetTypes = [Sun, Mercury, Venus, Earth, Mars, Jupiter, Saturn, Uranus, Neptune]

planetDataIO :: IO [PlanetData]
planetDataIO = do
  pics <- loadPictures imagePaths
  let combinedData = zip pics planetTypes
  return $ map (\(pic, pt) -> (pic, pt, lookupOrbitData pt)) combinedData

-- orbital data
lookupOrbitData :: PlanetType -> (Float, Float)
lookupOrbitData pt = case pt of
  Sun     -> (0, 0) -- Sun is at the center
  Mercury -> head orbitData
  Venus   -> orbitData !! 1
  Earth   -> orbitData !! 2
  Mars    -> orbitData !! 3
  Jupiter -> orbitData !! 4
  Saturn  -> orbitData !! 5
  Uranus  -> orbitData !! 6
  Neptune -> orbitData !! 7

orbitData :: [(Float, Float)]
orbitData = zip orbitRadii angularVelocities

update :: Float -> World -> World
update dt (time, planets) = (time + dt, planets)

render :: Maybe Picture -> World -> Picture
render maybeBackground (time, planetData) =   --render scene
  pictures $ maybeToList maybeBackground ++ map (drawPlanet time) planetData

calculatePosition :: Float -> Float -> Float -> Point
calculatePosition radius angularVelocity time = 
  (radius * cos (angularVelocity * time), radius * sin (angularVelocity * time))

maybeToList :: Maybe a -> [a]     -- Maybe -> list
maybeToList Nothing  = []
maybeToList (Just x) = [x]

-- Type definition: World state
type World = (Float, [PlanetData])

-- Planet data
type PlanetData = (Maybe Picture, PlanetType, (Float, Float))

-- Enum type: Planet type
data PlanetType = Sun | Mercury | Venus | Earth | Mars | Jupiter | Saturn | Uranus | Neptune
    deriving (Eq, Show)

drawPlanet :: Float -> PlanetData -> Picture
drawPlanet time (maybePic, planetType, (radius, angularVelocity)) =
  case maybePic of
    Just pic -> translate x y $ chooseScaleFunction planetType pic
    Nothing  -> blank
  where
    (x, y) = calculatePosition radius angularVelocity time

-- appropriate scaling function
chooseScaleFunction :: PlanetType -> Picture -> Picture
chooseScaleFunction pt = case pt of
  Sun     -> scaleSun
  Mercury -> scaleMercury
  Venus   -> scaleVenus
  Earth   -> scaleEarth
  Mars    -> scaleMars
  Jupiter -> scaleJupiter
  Saturn  -> scaleSaturn
  Uranus  -> scaleUranus
  Neptune -> scaleNeptune

-- scaling factor
calculateScaleFactor :: Float -> Float
calculateScaleFactor diameterRatio = diameterRatio * baseScaleFactor
  where baseScaleFactor = 4 

ignoreEvents :: Event -> world -> world
ignoreEvents _ world = world

loadPictures :: [FilePath] -> IO [Maybe Picture]
loadPictures paths = mapM loadJuicyPNG paths

--Calculating Angular Velocity:  thinking
-- Angular velocity can be based on the period of a planet's orbit around the Sun, measured in Earth years. 
---Since the Earth's orbital period is 1 year, its angular velocity can be set as 2π / 1 year.
-- The angular velocities of other planets can be calculated based on their orbital periods --
--(for example, Mercury is about 88 Earth days, Venus is about 224.7 Earth days, etc.), using the formula 2π / orbital period.

-- Orbital periods of planets
planetaryOrbitalPeriods :: [Float]
planetaryOrbitalPeriods = [0.40167, 1.025, 1.667, 3.135, 19.767, 49.1, 140.0167, 274.667]

-- Angular velocities per year
angularVelocities :: [Float]
angularVelocities = map (\period -> 2 * pi / period) planetaryOrbitalPeriods

-- Generic scaling function for planets
scalePlanet :: Float -> Picture -> Picture
scalePlanet diameterRatio pic = scale scaleFactor scaleFactor pic
  where scaleFactor = diameterRatio * globalScaleConstant

scaleSun :: Picture -> Picture
scaleSun = scalePlanet 109.2

scaleMercury :: Picture -> Picture
scaleMercury = scalePlanet 13.766

scaleVenus :: Picture -> Picture
scaleVenus = scalePlanet 14.898

scaleEarth :: Picture -> Picture
scaleEarth = scalePlanet 20.0

scaleMars :: Picture -> Picture
scaleMars = scalePlanet 14.064

scaleJupiter :: Picture -> Picture
scaleJupiter = scalePlanet 40.3

scaleSaturn :: Picture -> Picture
scaleSaturn = scalePlanet 36.9

scaleUranus :: Picture -> Picture
scaleUranus = scalePlanet 26.02

scaleNeptune :: Picture -> Picture
scaleNeptune = scalePlanet 25.76